// Software for Inference and Robotic Mapping with the Arak Process
// Copyright (C) 2005 Mark A. Paskin
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//
// Author contact information:
//   Mark A. Paskin
//   http://paskin.org
//   mark@paskin.org

#include <string>
#include <iostream>
#include <istream>
#include <ostream>
#include <fstream>

#include "global.hpp"
#include "jet.hpp"
#include "parsecl.hpp"

/**
 * A simple program to convert a grid of values to a portable pixmap
 * file (PPM) visualization (using the Matlab "jet" colormap).
 */
int main(int argc, char** argv) {
  using namespace Arak::Util;
  // Parse the command line.
  CommandLine cl;
  CommandLine::Parameter<double> 
    max("-min", "--min-intensity", "the surface value mapped to dark blue", 0.0);
  cl.add(max);
  CommandLine::Parameter<double> 
    min("-max", "--max-intensity", "the surface value mapped to dark red", 1.0);
  cl.add(min);
  CommandLine::Parameter<std::string> 
    inputPath("-i", "--input", 
	      "the input file of point color estimates", std::string());
  cl.add(inputPath);
  CommandLine::Parameter<std::string> 
    outputPath("-o", "--output", "the output PPM file", std::string());
  cl.add(outputPath);
  if (!cl.parse(argc, argv, std::cerr)) {
    cl.printUsage(std::cerr);
    exit(1);
  }

  // Grab the input and output streams.
  std::istream* in = &std::cin;
  if (inputPath.supplied()) {
    in = new std::ifstream(inputPath.value().data());
    if (!in->good()) {
      std::cerr << "Error: cannot open file " << inputPath.value()
		<< " for reading." << std::endl;
      exit(1);
    }
  }
  std::ostream* out = &std::cout;
  if (outputPath.supplied()) {
    out = new std::ofstream(outputPath.value().data());
    if (!out->good()) {
      std::cerr << "Error: cannot open file " << outputPath.value()
		<< " for writing." << std::endl;
      exit(1);
    }
  }

  // Skip the number of samples.
  double tmp;
  *in >> tmp;
  // Read the dimensions.
  int rows, cols;
  *in >> rows;
  *in >> cols;
  // Skip the bounding box.
  *in >> tmp >> tmp >> tmp >> tmp;

  // Compute the interpolated values and output the file.
  *out << "P3" << std::endl;                       // magic number
  *out << "# " << outputPath.value() << std::endl; // filename
  *out << cols << " " << rows << std::endl;        // resolution
  *out << 255 << std::endl;        // max-value
  for (int i = 0; i < rows; i++) {
    for (int j = 0; j < cols; j++) {
      double value;
      *in >> value;
      double intensity = 
	(value - min.value()) / (max.value() - min.value());
      intensity = (intensity < 0.0) ? 0.0 : intensity;
      intensity = (intensity > 1.0) ? 1.0 : intensity;
      int index = int(round((double(Arak::jet_size) - 1) * (1.0 - intensity)));
      // Compute the pixel value.
      *out << int(Arak::jet_rgb[index][0]) << " " 
	   << int(Arak::jet_rgb[index][1]) << " " 
	   << int(Arak::jet_rgb[index][2]) << std::endl; 
    }
    *out << std::endl;
  }
  out->flush();  

  // Delete any allocated streams.
  if (inputPath.supplied()) delete in;
  if (outputPath.supplied()) delete out;
  
  return 0;
}
